package dev.fyloz.colorrecipesexplorer.service

import dev.fyloz.colorrecipesexplorer.exception.RestException
import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.repository.RecipeStepRepository
import org.springframework.http.HttpStatus
import org.springframework.stereotype.Service

interface RecipeStepService : ModelService<RecipeStep, RecipeStepRepository> {
    /**
     *  Validates if the steps of the given [groupInformation] obey the following criteria:
     *
     *      * The position of the steps is greater or equals to 1
     *      * Each position is unique in the collection
     *      * There is no gap between positions
     *
     *  If any of those criteria are not met, an [InvalidStepsPositionsException] will be thrown.
     */
    fun validateGroupInformationSteps(groupInformation: RecipeGroupInformation)
}

@Service
class RecipeStepServiceImpl(recipeStepRepository: RecipeStepRepository) :
  AbstractModelService<RecipeStep, RecipeStepRepository>(recipeStepRepository),
  RecipeStepService {
    override fun idNotFoundException(id: Long) = recipeStepIdNotFoundException(id)
    override fun idAlreadyExistsException(id: Long) = recipeStepIdAlreadyExistsException(id)

    override fun validateGroupInformationSteps(groupInformation: RecipeGroupInformation) {
        val steps = groupInformation.steps
        val group = groupInformation.group

        if (steps == null) return

        val sortedSteps = steps.sortedBy { it.position }
        val errors = mutableSetOf<InvalidStepsPositionsError>()

        // Check if the first step position is 1
        fun isFirstStepPositionInvalid() =
          sortedSteps[0].position != 1

        // Check if any position is duplicated
        fun getDuplicatedPositionsErrors() =
          sortedSteps
            .groupBy { it.position }
            .filter { it.value.count() > 1 }
            .map { duplicatedStepsPositions(it.key, group) }

        // Check if there is any gap between steps positions
        // Knowing that steps are sorted by position, if the position of the step is not index + 1, there is a gap between them
        fun hasGapBetweenPositions() =
          sortedSteps
            .filterIndexed { index, step -> step.position != index + 1 }
            .isNotEmpty()

        // Find all errors and throw if there is any
        if (isFirstStepPositionInvalid()) errors += invalidFirstStepPosition(sortedSteps[0])
        errors += getDuplicatedPositionsErrors()
        if (errors.isEmpty() && hasGapBetweenPositions()) errors += gapBetweenStepsPositions(group)
        if (errors.isNotEmpty()) {
            throw InvalidStepsPositionsException(group, errors)
        }
    }
}

data class InvalidStepsPositionsError(
  val type: String,
  val details: String
)

class InvalidStepsPositionsException(
  val group: EmployeeGroup,
  val errors: Set<InvalidStepsPositionsError>
) :
  RestException(
    "invalid-recipestep-position",
    "Invalid steps position",
    HttpStatus.BAD_REQUEST,
    "The position of steps for the group ${group.name} are invalid",
    mapOf(
      "group" to group.name,
      "groupId" to group.id!!,
      "invalidSteps" to errors
    )
  )

const val INVALID_FIRST_STEP_POSITION_ERROR_CODE = "first"
const val DUPLICATED_STEPS_POSITIONS_ERROR_CODE = "duplicated"
const val GAP_BETWEEN_STEPS_POSITIONS_ERROR_CODE = "gap"

private fun invalidFirstStepPosition(
  step: RecipeStep
) = InvalidStepsPositionsError(
  INVALID_FIRST_STEP_POSITION_ERROR_CODE,
  "The position ${step.position} is under the minimum of 1"
)

private fun duplicatedStepsPositions(
  position: Int,
  group: EmployeeGroup
) = InvalidStepsPositionsError(
  DUPLICATED_STEPS_POSITIONS_ERROR_CODE,
  "The position $position is duplicated in the group ${group.name}"
)

private fun gapBetweenStepsPositions(
  group: EmployeeGroup
) = InvalidStepsPositionsError(
  GAP_BETWEEN_STEPS_POSITIONS_ERROR_CODE,
  "The positions for the steps of the group ${group.name} have gaps between them"
)
